#include <QtGui>
#include <QtNetwork>

class FortuneClient : public QObject
{
	Q_OBJECT
	Q_ENUMS(State)
	public:
		enum State { Disconnected, Connecting, WaitingTopics, Connected, Getting };
		FortuneClient(QObject *parent = NULL);
		~FortuneClient();
		void fConnect(const QString& host, quint16 port);
		void disconnect();
		State state() const;
		void get(const QString& topic);
		QStringList topics;
	signals:
		void stateChanged(FortuneClient::State state);
		void fortuneArrived(QString fortune);
		void error(QString description);
	private slots:
		void sockError(QAbstractSocket::SocketError socketError);
		void sockConnected();
		void sockReadyRead();
	private:
		void setState(State state);
		QTcpSocket sock;
		State m_state;
		QByteArray buf;
};

FortuneClient::FortuneClient(QObject *parent): QObject(parent)
{
	m_state = Disconnected;
	connect(&sock, SIGNAL(connected()), this, SLOT(sockConnected()));
	connect(&sock, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(sockError(QAbstractSocket::SocketError)));
	connect(&sock, SIGNAL(readyRead()), this, SLOT(sockReadyRead()));
};

void FortuneClient::fConnect(const QString& host, quint16 port)
{
	setState(Connecting);
	sock.connectToHost(host, port);
}

FortuneClient::State FortuneClient::state() const
{
	return m_state;
}

void FortuneClient::get(const QString& topic)
{
	if (state() != Connected)
	{
		emit error("Not connected");
		return;
	}
	if (!topics.contains(topic))
	{
		emit error("No such topic");
		return;
	}
	setState(Getting);
	sock.write((topic + "\n").toUtf8());
}

void FortuneClient::sockConnected()
{
	setState(WaitingTopics);
}

void FortuneClient::sockReadyRead()
{
	if (state() == WaitingTopics)
	{
		QByteArray cur = sock.readAll();
		buf.append(cur);
		if (cur.endsWith('\n'))
		{
			buf.resize(buf.size() - 1);
			topics = QString::fromUtf8(buf).split(",");
			buf.clear();
			setState(Connected);
		}
	}
	else if (state() == Getting)
	{
		QByteArray cur = sock.readAll();
		buf.append(cur);
		if (cur.endsWith('\n'))
		{
			buf.resize(buf.size() - 1);
			QString fortune = QString::fromUtf8(buf);
			emit fortuneArrived(fortune);
			buf.clear();
			setState(Connected);
		}
	}
	else
	{
		emit error("Data arrived, but we don't want it");
	}
}

void FortuneClient::sockError(QAbstractSocket::SocketError socketError)
{
	emit error(QString("Socket error: %1").arg( 
		QAbstractSocket::staticMetaObject.enumerator(
			QAbstractSocket::staticMetaObject.indexOfEnumerator("SocketError")
		).valueToKey(socketError)));

	setState(Disconnected);
}

void FortuneClient::setState(FortuneClient::State state)
{
	m_state = state;
	emit stateChanged(m_state);
}

void FortuneClient::disconnect()
{
	if (state() == Connected)
	{
		sock.disconnect();
		setState(Disconnected);
	}
}
FortuneClient::~FortuneClient()
{
	disconnect();
}

class Form : public QWidget
{
	Q_OBJECT
	public:
		Form();

	private slots:
		void displayError(QString error);
		void cConnect();
		void cStateChanged(FortuneClient::State state);
		void get();
	private:
		QStackedLayout *layout;
		QWidget *connData, *fortune;
		QLineEdit *server, *port;
		QComboBox *topic;
		QTextEdit *fortuneText;
		QPushButton *getButton, *connectButton;
		FortuneClient c;
};

Form::Form(): QWidget(NULL)
{
	layout = new QStackedLayout;
	connData = new QWidget;
	fortune = new QWidget;
	server = new QLineEdit;
	port = new QLineEdit;
	topic = new QComboBox;
	fortuneText = new QTextEdit;
	QFormLayout *connectForm = new QFormLayout;
	connectButton = new QPushButton("Connect >");
	connectForm->addRow("Server", server);
	connectForm->addRow("Port", port);
	connectForm->addRow(connectButton);
	QVBoxLayout *fortuneLayout = new QVBoxLayout;
	fortuneLayout->addWidget(topic);
	getButton = new QPushButton("Get");
	fortuneLayout->addWidget(getButton);
	fortuneLayout->addWidget(fortuneText);
	fortuneText->setReadOnly(true);
	connData->setLayout(connectForm);
	fortune->setLayout(fortuneLayout);
	connect(connectButton, SIGNAL(clicked()), this, SLOT(cConnect()));
	connect(getButton, SIGNAL(clicked()), this, SLOT(get()));
	connect(&c, SIGNAL(error(QString)), this, SLOT(displayError(QString)));
	connect(&c, SIGNAL(fortuneArrived(QString)), fortuneText, SLOT(setText(QString)));
	connect(&c, SIGNAL(stateChanged(FortuneClient::State)), this, SLOT(cStateChanged(FortuneClient::State)));
	
	layout->addWidget(connData);
	layout->addWidget(fortune);
	layout->setCurrentWidget(connData);
	setLayout(layout);
}

void Form::cConnect()
{
	getButton->setEnabled(false);
	c.fConnect(server->text(), port->text().toInt());
}

void Form::get()
{
	c.get(topic->currentText());
}

void Form::cStateChanged(FortuneClient::State state)
{
	if (state == FortuneClient::Connected)
	{
		if (!topic->count())
		{
			for (QStringList::iterator i = c.topics.begin(); i != c.topics.end(); ++i)
			{
				topic->addItem(*i);
			}
			layout->setCurrentWidget(fortune);
			getButton->setEnabled(true);
		}
	}
	else if (state == FortuneClient::Disconnected)
	{
		connectButton->setEnabled(true);
		layout->setCurrentWidget(connData);
	}
	else if (state == FortuneClient::Connecting)
	{
		connectButton->setEnabled(false);
	}
}

void Form::displayError(QString err)
{
	QMessageBox::critical(this, "Error", err);
}

int main(int ac, char** av)
{
	QApplication app(ac, av);
	Form f;
	f.show();
	return app.exec();
}

#include "main.moc"
